﻿using System;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Animation;
using System.Windows.Media.Media3D;

namespace Example05_3DAnimation01
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        // コンストラクター
        public MainWindow()
        {
            InitializeComponent();
            InitializeWindow();
        }

        private const double RADIUS = 10;       // Cube公転半径
        private const double TIME1 = 4;         // Cubeの自転周期
        private const double TIME2 = 8;         // Cubeの公転周期
        private AnimationClock _clockSR, _clockTrX, _clockTrZ;

        // [Avi]ボタンクリックイベント
        private void btnAvi_Click(object sender, RoutedEventArgs e)
        {
            var filepath = GetAviFilePath();
            if (0 < filepath.Length)
            {
                SaveAsAvi(filepath);
                MessageBox.Show("OK");
            }
        }

        // [開始]ボタンクリックイベント
        private void btnStart_Click(object sender, RoutedEventArgs e)
        {
            _clockSR.Controller.Begin();
            _clockTrX.Controller.Begin();
            _clockTrZ.Controller.Begin();
        }

        // [停止]ボタンクリックイベント
        private void btnStop_Click(object sender, RoutedEventArgs e)
        {
            _clockSR.Controller.Stop();
            _clockTrX.Controller.Stop();
            _clockTrZ.Controller.Stop();
        }


        // Cube各面のGeometryModel3Dを返すメソッド
        private GeometryModel3D GModel3D(Brush brush, Point3DCollection vp3col)
        {
            var meshgm3d = new MeshGeometry3D();
            meshgm3d.Positions = vp3col;
            meshgm3d.TriangleIndices = new Int32Collection(new int[] { 0, 1, 2, 0, 2, 3 });
            meshgm3d.TextureCoordinates = new PointCollection(new Point[] { 
                new Point(0, 0), new Point(0, 1), new Point(1, 1), new Point(1, 0) });

            var geometry = new GeometryModel3D();
            geometry.Geometry = meshgm3d;
            geometry.Material = new DiffuseMaterial(brush);

            return geometry;
        }

        // Aviファイル保存パス取得
        private string GetAviFilePath()
        {
            var dlg = new Microsoft.Win32.SaveFileDialog();
            dlg.Filter = "aviファイル|*.avi;*.avi";
            var path = String.Empty;

            if (dlg.ShowDialog() == true)
            {
                path = dlg.FileName;
            }

            return path;
        }

        // Cubeを生成するメソッド
        private void CreateCube()
        {
            // Cube六面の形状指定
            Point3DCollection[] p3col = new Point3DCollection[6];            
            p3col[0] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,-1,-1 1,-1,1 -1,-1,1 -1,-1,-1")));    // 底面
            p3col[1] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,-1 -1,1,1 1,1,1 1,1,-1")));        // 上面
            p3col[2] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,-1 -1,-1,-1 -1,-1,1 -1,1,1")));    // 左横面
            p3col[3] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,1,1 1,-1,1 1,-1,-1 1,1,-1")));        // 右横面
            p3col[4] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,1,-1 1,-1,-1 -1,-1,-1 -1,1,-1")));    // 奥
            p3col[5] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,1 -1,-1,1 1,-1,1 1,1,1")));        // 手前

            // Cube六面の色指定
            Color[] color = { Colors.Yellow, Colors.ForestGreen, Colors.MediumVioletRed,
                              Colors.DodgerBlue, Colors.Chocolate, Colors.Red};

            var groupM = new Model3DGroup();
            for (int i = 0; i < 6; i++)
            {
                var gmodel = GModel3D(new SolidColorBrush(color[i]), p3col[i]);
                groupM.Children.Add(gmodel);
            }            

            // Cubeの生成
            var cube = new ModelUIElement3D();
            cube.Model = groupM;

            // CubeのTransform設定
            var groupT = new Transform3DGroup();
            cube.Transform = groupT;

            // Cube自転のTransformとAnimation
            var axis3d = new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0);
            groupT.Children.Add(new RotateTransform3D(axis3d));
            var da1 = new DoubleAnimation(0, 360, new Duration(TimeSpan.FromSeconds(TIME1)));
            da1.RepeatBehavior = RepeatBehavior.Forever;
            _clockSR = da1.CreateClock();
            axis3d.ApplyAnimationClock(AxisAngleRotation3D.AngleProperty, _clockSR);
            _clockSR.Controller.Stop();

            // Cube公転のTransformとAnimation(X方向)
            var eg = new EllipseGeometry(new Point(0, 0), RADIUS , RADIUS);   // 公転軌道
            var tr3dX = new TranslateTransform3D();
            groupT.Children.Add(tr3dX);
            var da2 = new DoubleAnimationUsingPath();
            da2.Duration = new Duration(TimeSpan.FromSeconds(TIME2));
            da2.PathGeometry = PathGeometry.CreateFromGeometry(eg);
            da2.Source = PathAnimationSource.X;
            da2.RepeatBehavior = RepeatBehavior.Forever;
            _clockTrX = da2.CreateClock();
            tr3dX.ApplyAnimationClock(TranslateTransform3D.OffsetXProperty, _clockTrX);
            _clockTrX.Controller.Stop();

            // Cube公転のTransformとAnimation(Z方向)
            var tr3dZ = new TranslateTransform3D();
            groupT.Children.Add(tr3dZ);
            var da3 = new DoubleAnimationUsingPath();
            da3.Duration = new Duration(TimeSpan.FromSeconds(TIME2));
            da3.PathGeometry = PathGeometry.CreateFromGeometry(eg);
            da3.Source = PathAnimationSource.Y;
            da3.RepeatBehavior = RepeatBehavior.Forever;
            _clockTrZ = da3.CreateClock();
            tr3dZ.ApplyAnimationClock(TranslateTransform3D.OffsetZProperty, _clockTrZ);
            _clockTrZ.Controller.Stop();

            // Viewport3DへのCube配置
            myViewport.Children.Add(cube);
        }
        
        private void InitializeWindow()
        {
            CreateCube();            
        }

        // Aviファイル作成
        private void SaveAsAvi(string filepath)
        {
            var fps = 48;                            // フレームレート
            var totalFrames = (int)(fps * TIME2);    // 全フレーム数
            var secs = Enumerable.Range(0, totalFrames).Select(t => (((double)t) / fps));
            var aviManager = new AviFile.AviManager(filepath, false);
            AviFile.VideoStream aviStream = null;
            var tempFile = "frame.png";

            foreach (var sec in secs)
            {
                _clockSR.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                          TimeSeekOrigin.BeginTime);
                _clockTrX.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                           TimeSeekOrigin.BeginTime);
                _clockTrZ.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                           TimeSeekOrigin.BeginTime);
                cvs1.UpdateLayout();
                cvs1.SaveImage(tempFile);
                var bmp = new System.Drawing.Bitmap(tempFile);
                if (aviStream == null)
                    aviStream = aviManager.AddVideoStream(true, fps, bmp);
                else
                    aviStream.AddFrame(bmp);
                bmp.Dispose();
            }
            aviManager.Close();

            _clockSR.Controller.Stop();
            _clockTrX.Controller.Stop();
            _clockTrZ.Controller.Stop();

            System.IO.File.Delete(tempFile);
        }

    }

    public static class CanvasExtension
    {
        /// <summary> [SaveImage]：Canvasの画像をファイルに保存するCanvasクラスの拡張メソッド 
        /// 出典：http://qiita.com/hugo-sb/items/894914f6bbe224a45d49
        /// </summary>
        /// <param name="canvas"></param>
        /// <param name="path">保存する画像ファイル名</param>
        /// <param name="encoder">BitmapEncoderが指定されなかった場合はPNG形式になる</param>
        public static void SaveImage(this Canvas canvas, string path, BitmapEncoder encoder = null)
        {
            // レイアウトを再計算させる
            var size = new Size(canvas.ActualWidth, canvas.ActualHeight);
            canvas.Measure(size);
            canvas.Arrange(new Rect(size));

            // VisualObjectをBitmapに変換する
            var renderBitmap = new RenderTargetBitmap((int)size.Width,       // 画像の幅
                                                      (int)size.Height,      // 画像の高さ
                                                      96.0d,                 // 横96.0DPI
                                                      96.0d,                 // 縦96.0DPI
                                                      PixelFormats.Pbgra32); // 32bit(RGBA各8bit)
            renderBitmap.Render(canvas);

            // 出力用の FileStream を作成する
            using (var stream = new System.IO.FileStream(path, System.IO.FileMode.Create))
            {
                // 変換したBitmapをエンコードしてFileStreamに保存する。
                // BitmapEncoder が指定されなかった場合は、PNG形式とする。
                encoder = encoder ?? new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(renderBitmap));
                encoder.Save(stream);
            }
        }

    }

}
